/*********************************************************************************************/
/*                                                                                           */
/*                                                                File: NeuronController.cpp */
/*                                                                                           */
/*     Handles the GPIO pins and sends pager messages when appropriate                       */
/*     ===============================================================                       */
/*                                                                                           */
/*     V0.1     23-MAY-2017      Te   Created out of Controller.cpp and neuronspi.c          */
/*     V0.3     06-JUL-2017      Te   Send an alarm every hour while the input is active     */
/*                                                                                           */
/*********************************************************************************************/

/*
 *  Controller for Swissphone TPS 
 *  Copyright (C) 2017 P. Tellenbach
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/spi/spidev.h>
#include <sys/mman.h>

#include "configuration.h"
#include "ipcmessage.h"
#include "Logger.h"
#include "armspi.h"

#define VERSION "NeuronController V0.02"

// Define the register offsets
#define GPLEVEL         13
#define GPPUD           37
#define GPPUDCLK        38

// Define the bit masks for the pins
#define GPIO04   0x0000010
#define GPIO07   0x0000080
#define GPIO11   0x0000800
#define GPIO17   0x0020000
#define GPIO22   0x0400000
#define GPIO23   0x0800000
#define GPIO24   0x1000000
#define GPIO27   0x8000000

typedef struct
{
   bool   active  ;
   int    counter ;
   time_t timestamp ;
} INPUTSTATUS ;

static INPUT       inputs[INPUT_COUNT] ;
static INPUTSTATUS status[INPUT_COUNT] ;
static IPCMESSAGE  message ;
static int         queue ;

static arm_handle  handle ;

/************************************/
 static bool ReadInputConfiguration()
/************************************/

{
   // Clear previous configuration
   memset( inputs, 0, sizeof(inputs) ) ;

   FILE *file = fopen( CONFIG_FILE, "r" ) ;

   if( file == NULL )
      return false ;

   int   section = S_NONE ;
   int   index   = -1 ;
   char  line[512] ;
   int   pi ;

   while( fgets(line,sizeof(line),file) != NULL )
   {
      if( strcmp(line,"[input]\n") == 0 )
      {
         if( section != S_INPUT )
         {
            section = S_INPUT ;
            index   = 0 ;
         }
         else
            ++index ;

         pi = 0 ;
      }
      else if( (section == S_INPUT) && (index >= 0) && (index < INPUT_COUNT) )
      {
         INPUT &input = inputs[index] ;

         if( (strncmp(line,"on=",3) == 0) )
            GetString( line + 3, input.on, sizeof(input.on) ) ;
         else if( (strncmp(line,"off=",4) == 0) )
            GetString( line + 4, input.off, sizeof(input.off) ) ;
         else if( (strncmp(line,"from=",5) == 0) )
            GetString( line + 5, input.from, sizeof(input.from) ) ;
         else if( (strncmp(line,"to=",3) == 0) )
            GetString( line + 3, input.to, sizeof(input.to) ) ;
         else if( (strncmp(line,"debounce=",9) == 0) )
            input.debounce = atoi( line + 9) / 100 ;
         else if( (strncmp(line,"number=",7) == 0) && (pi >= 0) && (pi < PAGER_COUNT) )
            GetString( line + 7, input.pager[pi].number, sizeof(input.pager[0].number) ) ;
         else if( (strncmp(line,"legitimation=",13) == 0) && (pi >= 0) && (pi < PAGER_COUNT) )
            GetString( line + 13, input.pager[pi++].legi, sizeof(input.pager[0].legi) ) ;
      }
   }

   fclose( file ) ;
   return true ;
}

/***********************/
 static int InitNeuron()
/***********************/

{
    if( arm_init(&handle,"/dev/spidev0.1",12000000,0,NULL) == -1 )
       return -1 ;

    for( int i = 0; i < 5; i++ )
    {
        if( idle_op(&handle) != -1 )
           return 0 ;

        sleep( 1 ) ;
    }

    close( handle.fd ) ;
    return -1 ;
}

/******************/
 static void Wait()
/******************/

{
   struct timespec requested ;
   struct timespec remaining ;

   requested.tv_sec  = 0 ;
   requested.tv_nsec = 100000000 ;

   nanosleep( &requested, &remaining ) ;
}

/***********************/
 static int ReadInputs()
/***********************/

{
   uint16_t buffer[128];
   buffer[0] = 10;
   buffer[1] = 1;

   if( read_regs(&handle,0,10,buffer) == -1 )
   {
      WriteLog( "error: read_regs failed" ) ;

      int errors = 0 ;
      int limit  = 5 ;

      sleep( 1 ) ;

      buffer[0] = 10 ;
      buffer[1] =  1 ;
      while( read_regs(&handle,0,10,buffer) == -1 )
      {
         if( ++errors > limit )
         {
            WriteLog( "error: read_regs still fails" ) ;

            errors = 0 ;
            limit *= 2 ;
         }

         buffer[0] = 10 ;
         buffer[1] =  1 ;
      }
   }

   return buffer[0] ;
}

/**********************************************/
 static void HandleChange( int index, bool on )
/**********************************************/

{
   char line[120] ;
   sprintf( line, "Eingang %d %s ", index + 1, on ? "aktiv  " : "inaktiv" ) ;

   const char *text = on ? inputs[index].on : inputs[index].off ;

   if( strlen(text) == 0 )
   {
      strcat( line, "Der Text ist leer" ) ;
      WriteLog( line ) ;
      return ;
   }

   INPUT &input = inputs[index] ;
   int count    = 0 ;

   for( int i = 0; i < PAGER_COUNT; i++ )
   {
      if( strlen(input.pager[i].number) > 0 )
         ++count ;
   }

   if( count == 0 )
   {
      strcat( line, "Keine Pager definiert" ) ;
      WriteLog( line ) ;
      return ;
   }

   if( strlen(input.from) > 0 )
   {
      time_t     now = time( NULL ) ;
      struct tm *pt  = localtime( &now ) ;

      char str[12] ;
      sprintf( str, "%02d:%02d", pt->tm_hour, pt->tm_min ) ;

      if( (strcmp(input.from,str) > 0) || (strcmp(input.to,str) < 0) )
      {
         strcat( line, "Nicht im Zeitfenster" ) ;
         WriteLog( line ) ;
         return ;
      }
   }

   message.type    = 1 ;
   message.command = CMD_PAGE ;
   strcpy( message.text, text ) ;

   strcat( line, "Meldung an:" ) ;
   for( int i = 0; i < PAGER_COUNT; i++ )
   {
      message.pager[i] = input.pager[i] ;

      const char *number = input.pager[i].number ;

      if( strlen(number) > 0 )
      {

         strcat( line, " " ) ;
         strcat( line, number ) ;
      }
   }

   msgsnd( queue, &message, sizeof(message) - sizeof(long), IPC_NOWAIT ) ; 

   WriteLog( line ) ;
}

/*************************************/
 static void UpdateStatus( int state )
/*************************************/

{
   time_t now    = time( NULL ) ;
   bool   active = false ;
   int    mask   = 1 ;

   for( int i = 0; i < INPUT_COUNT; i++ )
   {
      active = (state & mask) != 0 ;

      if( status[i].active == active )
      {
         status[i].counter = 0 ;

         if( active && (status[i].timestamp + 3600 < now) )
         {
            status[i].timestamp = now ;
            HandleChange( i, active ) ;
         }
      }
      else
      {
         status[i].counter += 1 ;

         if( status[i].counter >= inputs[i].debounce )
         {
            status[i].timestamp = now ;
            status[i].active    = active ;
            status[i].counter   = 0 ;
            HandleChange( i, active ) ;
         }
      }

      mask <<= 1 ;
   }
}

/***********/
 int main( )
/***********/
{
   SetLogfile( "/home/pi/controller.log" ) ;

   WriteLog( VERSION ) ;

   // Read the configuration
   if( !ReadInputConfiguration() )
   {
      WriteLog( "error: can't read configuration" ) ;
      return 1 ;
   }

   // Create the queue
   queue = msgget( 3063, 0666 | IPC_CREAT ) ;

   if( queue == -1 )
   {
      WriteLog( "error: can't create queue" ) ;
      return 2 ;
   }

   if( InitNeuron() == -1 )
   {
      WriteLog( "error: Neuron init failed" ) ;
      return 3 ;
   }

   int loopcounter = 0 ;

   for( ;; )
   {
      if( ++loopcounter > 10 )
      {
         if( msgrcv(queue,&message,sizeof(message)-sizeof(long),2,IPC_NOWAIT) != -1 )
         {
            ReadInputConfiguration() ;
            WriteLog( "Reload configuration" ) ;
         }
      }

      Wait() ;

      UpdateStatus( ReadInputs() ) ;
   }

   // Should be unreachable ...
   close( handle.fd ) ;
   
   return 0 ;
}
