/******************************************************************************/
/*                                                                            */
/*                                                              FILE: Pager.c */
/*                                                                            */
/*        Sends pager messages to a swissphone pager                          */
/*        ==========================================                          */
/*                                                                            */
/*        V0.01   07-MAR-2017  Te                                             */
/*        V0.10   20-MAR-2017  Te                                             */
/*        V0.20   21-APR-2017  Te                                             */
/*        V0.21   07-JUL-2017  Te   Added GPL license                         */
/*                                                                            */
/******************************************************************************/

/*
 *  Sends pager messages for Swissphone TPS 
 *  Copyright (C) 2017 P. Tellenbach
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#include "Swissphone.h"
#include "configuration.h"
#include "ipcmessage.h"
#include "Logger.h"

#define VERSION "SwissphonePager V0.60"

#define PORT             "1234"

#define INTERVAL             60
#define DELAY    (INTERVAL + 2)

static Swissphone swissphone ;
static SERVER     server  ;
static IPCMESSAGE message ;

static time_t timestamp      = 0 ;
static int    messagecounter = 0 ;

/*****************************************************/
 static bool ReadServerConfiguration( SERVER &server )
/*****************************************************/

{
   FILE *file = fopen( CONFIG_FILE, "r" ) ;

   if( file == NULL )
      return false ;

   char line[512] ;

   while( fgets(line,sizeof(line),file) != NULL )
   {
      if( strncmp(line,"ip=",3) == 0 )
         GetString( line + 3, server.ip, sizeof(server.ip) ) ;
      else if( strncmp(line,"user=",5) == 0 )
         GetString( line + 5, server.user, sizeof(server.user) ) ;
      else if( strncmp(line,"password=",9) == 0 )
         GetString( line + 9, server.password, sizeof(server.password) ) ;
      else if( strncmp(line,"limit=",6) == 0 )
         server.limit = atoi( line + 6 ) ;
   }

   fclose( file ) ;

   return (strlen(server.ip) > 0) && (strlen(server.user) > 0) && (strlen(server.password) > 0) ;
}

/**********************************************************/
 static int SendPage( SERVER &server, IPCMESSAGE &message )
/**********************************************************/

{
   time_t now = time( NULL ) ;

   if( now > timestamp + INTERVAL )
      messagecounter = 0 ;
   else
   {
      if( messagecounter >= server.limit )
      {
         sleep( DELAY ) ;
         messagecounter = 0 ;
      }
   }

   int index = 0 ;

   while( index < PAGER_COUNT )
   {
      if( !swissphone.Connect(server.ip,PORT) )
         return SE_CONNECT ;

      int result = swissphone.Login( server.user, server.password ) ;

      if( result != 0 )
      {
         swissphone.Disconnect() ;
         return result ;
      }

      timestamp = time( NULL ) ;

      if( !swissphone.CenterReady() )
      {
         swissphone.Disconnect() ;
         return 205 ;
      }

      while( index < PAGER_COUNT )
      {
         PAGER &pager = message.pager[index++] ;

         if( strlen(pager.number) > 0 )
         {
            result = swissphone.SendPage( pager, message.text ) ;

            WriteLog( "%s -> %d", pager.number, result ) ;

            if( ++messagecounter >= server.limit )
               break ;
         }
      }

      swissphone.Logout() ;
      swissphone.Disconnect() ;

      if( messagecounter >= server.limit )
      {
         sleep( DELAY ) ;
         messagecounter = 0 ;
      }
   }

   return 0 ;
}

/**********************************/
 int main( int argc, char *argv[] )
/**********************************/

{
   SetLogfile( "/home/pi/pager.log" ) ;

   WriteLog( VERSION ) ;

   if( !ReadServerConfiguration(server) )
   {
      WriteLog( "error: can't read configuration" ) ;
      return SE_CONFIG ;
   }

   int queue = msgget( 3063, 0666 | IPC_CREAT ) ;

   if( queue == -1 )
   {
      WriteLog( "error: can't create queue" ) ;
      return SE_QUEUE ;
   }

   while( msgrcv(queue,&message,sizeof(message)-sizeof(long),1,0) != -1 )
   {
      
      switch( message.command )
      {
       case CMD_PAGE   : WriteLog( "Meldung: \"%s\"", message.text ) ;
                         SendPage( server, message ) ;
                         break ;
       case CMD_RELOAD : WriteLog( "Reload-Befehl" ) ;
                         break ;
       case CMD_EXIT   : WriteLog( "Exit-Befehl" ) ;
                         return 0 ;
      }
   }

   return SE_RECEIVE ;
}
