/******************************************************************************/
/*                                                                            */
/*                                                       FILE: Swissphone.cpp */
/*                                                                            */
/*        Implements the TAP protocol for the Swissphone paging center        */
/*        ============================================================        */
/*                                                                            */
/*        V0.01   20-MAR-2017  Te                                             */
/*        V0.02   07-JUL-2017  Te   Added GPL license                         */
/*                                                                            */
/******************************************************************************/

/*
 *  A class which implements the TAP protocol for Swissphone TPS 
 *  Copyright (C) 2017 P. Tellenbach
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <errno.h>
#include <ctype.h>
#include <stdlib.h>

#include "Swissphone.h"
#include "Logger.h"

/**************************************/
 static int MatchID( const char *line )
/**************************************/

{
   if( strstr(line,"ID=\r") != NULL )
      return 1 ;

   return 0 ;
}

/*****************************************/
 static int MatchLogin( const char *line )
/*****************************************/

{
   if( strstr(line,"\r\06\rFRZ II\r") != NULL )
      return 1 ;

   if( strstr(line,"\r\25\r") != NULL )
      return 2 ;

   return 0 ;
}

/******************************************/
 static int MatchAccept( const char *line )
/******************************************/

{
   if( strstr(line,"\r\06\r") != NULL )
      return 1 ;

   if( strstr(line,"\r\25\r") != NULL )
      return 2 ;

   return 0 ;
}

/******************************************************************************/
 static void BuildMessage( char *buffer, const PAGER &pager, const char *text )
/******************************************************************************/

{
   const char *pn = pager.number ;
   const char *pl = pager.legi ;
   char       *ps = buffer ;

   *ps++ = '\x02' ;

   for( int i = 0; (i < NUMBER_LENGTH) && (*pn != '\0'); i++ )
      *ps++ = *pn++ ;

   *ps++ = '\r' ;
   
   for( int i = 0; (i < 160) && (*text != '\0'); i++ )
   {
      if( (*text & 0x80) == 0 )
      {
         if( *text < ' ' )
            *ps = ' ' ;
         else
            *ps = *text ;
      }
      else if( (*text & 0xFF) == 0xC3 )
      {
         if( text[1] != '\0' )
            ++text ;

         switch( *text & 0xFF )
         {
          case 0xA4 : *ps = 0x7B ; break ;
          case 0xB6 : *ps = 0x7C ; break ;
          case 0xBC : *ps = 0x7D ; break ;
          case 0x84 : *ps = 0x7B ; break ;
          case 0x96 : *ps = 0x7C ; break ;
          case 0x9C : *ps = 0x7D ; break ;
          case 0xBB : *ps = 0x7E ; break ;
          case 0xA9 : *ps = 0x5B ; break ;
          case 0xA7 : *ps = 0x5C ; break ;
          case 0xAA : *ps = 0x5D ; break ;
          case 0xAE : *ps = 0x5E ; break ;
          case 0xA8 : *ps = 0x5F ; break ;
          case 0xB9 : *ps = 0x23 ; break ;
          case 0xA0 : *ps = 0x40 ; break ;
          case 0xB4 : *ps = 0x60 ; break ;
          default   : *ps = ' '  ; break ;
         }
      }
      else
         *ps = ' ' ;

      ++text ;
      ++ps ;
   }

   *ps++ = '\r'   ;

   if( strlen(pl) > 0 )
   {
      for( int i = 0; (i < LEGI_LENGTH) && (*pl != '\0'); i++ )
         *ps++ = *pl++ ;

      *ps++ = '\r' ;
   }

   *ps++ = '\x03' ;
   *ps   = '\0'   ;

   int checksum = 0 ;
   ps           = buffer ;

   while( *ps )
      checksum += *ps++ ;

   *ps++ = (((checksum >> 8) & 0x000F) + '0') ;
   *ps++ = (((checksum >> 4) & 0x000F) + '0') ;
   *ps++ = (( checksum       & 0x000F) + '0') ;

   *ps++ = '\r' ;
   *ps   = '\0' ;
}

/**************************************************************/
 bool Swissphone::Connect( const char *host, const char *port )
/**************************************************************/

{
   // Create a new socket
   sock = socket( AF_INET, SOCK_STREAM, 0 ) ;
   
   if( sock == -1 )
   {
      WriteLog( "error: socket failed" ) ;
      return false ;
   }
    
   // Bind it to the local address
   struct sockaddr_in local ;

   memset( &local, 0, sizeof(local) ) ;
   local.sin_family      = AF_INET ;
   local.sin_addr.s_addr = htonl( INADDR_ANY ) ;
   local.sin_port        = htons( 0 ) ;
    
   if( bind(sock,(struct sockaddr *)&local,sizeof(local)) == -1 )
   {
      WriteLog( "error: bind failed" ) ;
      close( sock ) ;
      return false ;
   }

   struct addrinfo  hints ;
   struct addrinfo *result ;

   memset( &hints, 0, sizeof(hints) ) ;
   hints.ai_family   = AF_INET ;
   hints.ai_socktype = SOCK_STREAM ;

   if( getaddrinfo(host,port,&hints,&result) != 0 )
   {
      WriteLog( "error: hostname resolution failed" ) ;
      close( sock ) ;
      return false ;
   }

   if( connect(sock,result->ai_addr,result->ai_addrlen) == -1 )
   {
      WriteLog( "error: connecting to %s failed", host ) ;
      freeaddrinfo( result ) ;
      close( sock ) ;
      return false ;
   }

   freeaddrinfo( result ) ;

   return true ;
}

/**************************************/
 int Swissphone::Receive( int seconds )
/**************************************/

{
   struct timeval tv ;
   fd_set         set ;

   FD_ZERO( &set ) ;
   FD_SET( sock, &set ) ;

   tv.tv_sec  = seconds ;
   tv.tv_usec = 0 ;

   int result = select( sock + 1, &set, NULL, NULL, &tv ) ;

   if( result == -1 )
   {
      WriteLog( "error: select failed" ) ;
      return -1 ;
   }

   if( result == 0 )
      return 0 ;

   int len = read( sock, buffer + index, sizeof(buffer) - 1 - index ) ;

   if( len == -1 )
   {
      WriteLog( "error: read failed" ) ;
      return -1 ;
   }

   if( len == 0 )
   {
      WriteLog( "error: server disconnected" ) ;
      return -1 ;
   }

   index += len ;

   buffer[index] = '\0' ;

   return 1 ;
}

/***************************************************************************/
 int Swissphone::CheckResponse( int seconds, int (*match)(const char *str) )
/***************************************************************************/

{
   time_t timeout = time( NULL ) + seconds ;
   int    result  = 0 ;

   index = 0 ;

   for( ;; )
   {
      time_t now = time( NULL ) ;

      if( now >= timeout )
         break ;

      switch( Receive(timeout-now) )
      {
       case -1: return 0 ;

       case  0: break ;

       case  1: result = match( buffer ) ;

                if( result != 0 )
                   return result ;
                break ;
      }
   }

   return 0 ;
}

/********************************/
 bool Swissphone::StartProtocol()
/********************************/

{
   for( int i = 0; i < 3; i++ )
   {
      index = 0 ;

      if( write(sock,"\r",1) != 1 )
      {
         WriteLog( "error: write failed" ) ;
         return false ;
      }

      if( CheckResponse(2,MatchID) == 1 )
         return true ;
   }

   WriteLog( "error: timeout while waiting for ID=" ) ;

   return false ;
}

/***************************************************************/
 int Swissphone::Login( const char *user, const char *password )
/***************************************************************/

{
   if( !StartProtocol() )
      return 203 ;

   sprintf( buffer, "\33PG1%s:%s\r", user, password ) ;

   if( write(sock,buffer,strlen(buffer)) != (ssize_t)strlen(buffer) )
   {
      WriteLog( "error: write failed" ) ;
      return false ;
   }

   switch( CheckResponse(2,MatchLogin) )
   {
    case 0: WriteLog( "error: login timeout" ) ;
            return 204 ;

    case 1: return 0 ;

    case 2: WriteLog( "error: invalid login" ) ;
            return 204 ;
   }

   return 204 ;
}

/******************************/
 bool Swissphone::CenterReady()
/******************************/

{
   time_t timeout = time( NULL ) + 60 ;

   for( ;; )
   {
      if( strstr(buffer,"\33[p\r") != NULL )
         return true ;

      time_t now = time( NULL ) ;

      if( now >= timeout )
         break ;

      if( Receive(timeout-now) == -1 )
         return false ;
   }

   WriteLog( "error: ready timeout" ) ;

   return false ;
}

/**************************************/
 const char *SkipWord( const char *ps )
/**************************************/

{
   while( (*ps != '\0') && (*ps != ' ') )
      ++ps ;

   return ps ;
}

/**************************/
 int Swissphone::Accepted()
/**************************/

{
   const char *ps ;

   switch( CheckResponse(20,MatchAccept) )
   {
    case 0: WriteLog( "error: message timeout" ) ;
            return 200 ;

    case 1: ps = SkipWord( buffer ) ;

            if( *ps++ != ' ' )
               return 201 ;

            return atoi( ps ) ;

    case 2: return atoi(buffer) ;
   }

   return 202 ;
}

/****************************************************************/
 int Swissphone::SendPage( const PAGER &pager, const char *text )
/****************************************************************/

{
   BuildMessage( buffer, pager, text ) ;

   write( sock, buffer, strlen(buffer) ) ;

   int result = Accepted() ;

   return result ;
}

/*************************/
 void Swissphone::Logout()
/*************************/

{
   write( sock, "\04\r", 2 ) ;

   index = 0 ;
   Receive( 2 ) ;
}

/*****************************/
 void Swissphone::Disconnect()
/*****************************/

{
   close( sock ) ;
   sock = -1 ;
}
